/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wKuttaResidual.h"
#include "wKuttaElement.h"

#include "wElement.h"
#include "wCache.h"
#include "wGauss.h"

using namespace tbox;
using namespace dart;

/**
 * @brief Build the residual matrix for a fixed-point iteration, on one Kutta element
 *
 * A = 1/S * \int (psi + grad_psi*v_inf) * v*grad_phi dS
 * NB: v_inf = [1, 0, 0]
 */
Eigen::MatrixXd KuttaResidual::buildFixed(KuttaElement const &ke, std::vector<double> const &phi)
{
    // Get shape functions and Gauss points
    Cache &sCache = ke.surE->getVCache();
    Gauss &sGauss = sCache.getVGauss();
    Eigen::MatrixXd const &dSfV = ke.volE->getVCache().getDsf(0);
    // Get Jacobian
    Eigen::MatrixXd const &iJ = ke.volE->getJinv(0);
    // Get size
    size_t nRow = ke.nRow;
    size_t nDim = ke.nDim;

    // Stabilization (disabled in 2D)
    Eigen::VectorXd dSfp = Eigen::VectorXd::Zero(nRow);
    if (nDim == 3)
    {
        Eigen::MatrixXd dSf(nRow, nDim);
        for (size_t i = 0; i < nRow; ++i)
            for (size_t j = 0; j < nDim; ++j)
                dSf(i, j) = dSfV(j, ke.vsMap.at(ke.teN[i].first));
        dSfp = 0.5 * sqrt(ke.surE->getVol()) * dSf * iJ.transpose() * Eigen::Vector3d(1, 0, 0).middleRows(0, nDim);
    }
    // Velocity
    Eigen::RowVectorXd Aup = 1 / ke.surE->getVol() * ke.volE->computeGradient(phi, 0).transpose() * iJ * dSfV;

    // Build
    Eigen::MatrixXd A = Eigen::MatrixXd::Zero(nRow, ke.nCol);
    for (size_t k = 0; k < sGauss.getN(); ++k)
    {
        // shape functions
        Eigen::VectorXd const &sf = sCache.getSf(k);
        Eigen::VectorXd sfp(nRow);
        for (size_t i = 0; i < nRow; ++i)
            sfp(i) = sf(ke.teN[i].first) + dSfp(i);
        // wake contribution
        A += ke.sign * sfp * Aup * sGauss.getW(k) * ke.surE->getDetJ(k);
    }
    return A;
}

/**
 * @brief Build the residual vector, on one Kutta element
 *
 * b = 1/S * \int (psi + grad_psi*v_inf) * (grad_phi)^2 dS
 * NB: v_inf = [1, 0, 0]
 */
Eigen::VectorXd KuttaResidual::build(KuttaElement const &ke, std::vector<double> const &phi)
{
    // Get shape functions and Gauss points
    Cache &sCache = ke.surE->getVCache();
    Gauss &sGauss = sCache.getVGauss();
    Eigen::MatrixXd const &dSfV = ke.volE->getVCache().getDsf(0);
    // Get Jacobian
    Eigen::MatrixXd const &iJ = ke.volE->getJinv(0);
    // Get size
    size_t nRow = ke.nRow;
    size_t nDim = ke.nDim;

    // Stabilization (disabled in 2D)
    Eigen::VectorXd dSfp = Eigen::VectorXd::Zero(nRow);
    if (nDim == 3)
    {
        Eigen::MatrixXd dSf(nRow, nDim);
        for (size_t i = 0; i < nRow; ++i)
            for (size_t j = 0; j < nDim; ++j)
                dSf(i, j) = dSfV(j, ke.vsMap.at(ke.teN[i].first));
        dSfp = 0.5 * sqrt(ke.surE->getVol()) * dSf * iJ.transpose() * Eigen::Vector3d(1, 0, 0).middleRows(0, nDim);
    }
    // Velocity
    double dPhi2 = 1 / ke.surE->getVol() * ke.volE->computeGradient(phi, 0).squaredNorm();

    // Build
    Eigen::VectorXd b = Eigen::VectorXd::Zero(nRow);
    for (size_t k = 0; k < sGauss.getN(); ++k)
    {
        // shape functions
        Eigen::VectorXd const &sf = sCache.getSf(k);
        Eigen::VectorXd sfp(nRow);
        for (size_t i = 0; i < nRow; ++i)
            sfp(i) = sf(ke.teN[i].first) + dSfp(i);
        // wake contribution
        b += ke.sign * sfp * dPhi2 * sGauss.getW(k) * ke.surE->getDetJ(k);
    }
    return b;
}

/**
 * @brief Build the gradient of the residual vector with respect to the flow variable (jacobian), on one Kutta element
 *
 * A = 1/S * \int (psi + grad_psi*v_inf) * 2*(grad_phi_u)*dgrad_phi_u dS
 * NB: v_inf = [1, 0, 0]
 */
Eigen::MatrixXd KuttaResidual::buildGradientFlow(KuttaElement const &ke, std::vector<double> const &phi)
{
    // Build
    Eigen::MatrixXd A = KuttaResidual::buildFixed(ke, phi);
    return 2 * A;
}

/**
 * @brief Build the gradient of the residual vector with respect to the nodes, on one Kutta element
 *
 * A = d( 1/S * \int (psi + grad_psi*v_inf) * (grad_phi)^2 dS )
 *   = d1/S * \int (psi + grad_psi*v_inf) * (grad_phi)^2 dS
 *   + 1/S * \int dgrad_psi*v_inf * (grad_phi)^2 dS
 *   + 1/S * \int (psi + grad_psi*v_inf) * d(grad_phi)^2 dS
 *   + 1/S * \int (psi + grad_psi*v_inf) * (grad_phi)^2 ddS
 * NB: v_inf = [1, 0, 0]
 */
std::tuple<Eigen::MatrixXd, Eigen::MatrixXd> KuttaResidual::buildGradientMesh(KuttaElement const &ke, std::vector<double> const &phi)
{
    // Get shape functions and Gauss points
    Cache &surCache = ke.surE->getVCache();
    Gauss &surGauss = surCache.getVGauss();
    Eigen::MatrixXd const &dSfV = ke.volE->getVCache().getDsf(0);
    // Get Jacobian
    Eigen::MatrixXd const &iJ = ke.volE->getJinv(0);
    std::vector<Eigen::MatrixXd> const &dJ = ke.volE->getGradJ(0);
    // Get surface and gradient
    double surf = ke.surE->getVol();
    std::vector<double> const &dSurf = ke.surE->getGradVol();
    // Get size
    size_t nRow = ke.nRow;
    size_t nDim = ke.nDim;
    size_t nCol = ke.nCol;
    size_t nSur = ke.surE->nodes.size();

    // Stabilization (disabled in 2D)
    Eigen::VectorXd gradSfp = Eigen::VectorXd::Zero(nRow);
    std::vector<Eigen::VectorXd> dGradSfV(nDim * nCol, Eigen::VectorXd::Zero(nRow));
    std::vector<Eigen::VectorXd> dGradSfS(nDim * nSur, Eigen::VectorXd::Zero(nRow));
    if (nDim == 3)
    {
        // term
        Eigen::VectorXd vi = Eigen::Vector3d(1, 0, 0).middleRows(0, nDim);
        double sqSurf = sqrt(surf);
        Eigen::MatrixXd dSf(nRow, nDim);
        for (size_t i = 0; i < nRow; ++i)
            for (size_t m = 0; m < nDim; ++m)
                dSf(i, m) = dSfV(m, ke.vsMap.at(ke.teN[i].first));
        gradSfp = 0.5 * sqSurf * dSf * iJ.transpose() * vi;
        // volume gradient
        for (size_t i = 0; i < nCol; ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                dGradSfV[idx] = 0.5 * sqSurf * dSf * (-iJ * dJ[idx] * iJ).transpose() * vi; // S * dgrad_psi
            }
        }
        // surface gradient
        for (size_t i = 0; i < nSur; ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                dGradSfS[idx] = 0.5 * 0.5 / sqSurf * dSurf[idx] * dSf * iJ.transpose() * vi; // dS * grad_psi
            }
        }
    }
    // Velocity and normalization
    Eigen::VectorXd dPhi = ke.volE->computeGradient(phi, 0);
    double nrm = 1 / surf;

    // Build
    Eigen::MatrixXd Av = Eigen::MatrixXd::Zero(nRow, nDim * nCol);
    Eigen::MatrixXd As = Eigen::MatrixXd::Zero(nRow, nDim * nSur);
    for (size_t k = 0; k < surGauss.getN(); ++k)
    {
        // weight and Jacobian determinant
        double w = surGauss.getW(k);
        double detJ = ke.surE->getDetJ(k);
        std::vector<double> dDetJ = ke.surE->getGradDetJ(k);
        // shape functions
        Eigen::VectorXd const &sf = surCache.getSf(k);
        Eigen::VectorXd sfp(nRow);
        for (size_t i = 0; i < nRow; ++i)
            sfp(i) = sf(ke.teN[i].first) + gradSfp(i);

        // Volume
        for (size_t i = 0; i < nCol; ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                // gradient of velocity jump
                Av.col(idx) += ke.sign * nrm * w * sfp * 2 * dPhi.transpose() * (-iJ * dJ[idx] * dPhi) * detJ; // 1/S * (psi + grad_psi) * dgrad_phi^2 * detJ
                // gradient of stabilization term (volume)
                Av.col(idx) += ke.sign * nrm * w * dGradSfV[idx] * dPhi.dot(dPhi) * detJ; // 1/S * dgrad_psi * grad_phi^2 * detJ
            }
        }
        // Surface
        for (size_t i = 0; i < nSur; ++i)
        {
            for (size_t m = 0; m < nDim; ++m)
            {
                size_t idx = i * nDim + m;
                // gradient of normalization
                As.col(idx) -= ke.sign * nrm * nrm * dSurf[idx] * w * sfp * dPhi.dot(dPhi) * detJ; // d1/S * (psi + grad_psi) * grad_phi^2 * detJ
                // gradient of stabilization term (surface)
                As.col(idx) += ke.sign * nrm * w * dGradSfS[idx] * dPhi.dot(dPhi) * detJ; // 1/S * dgrad_psi * grad_phi^2 * detJ
                // gradient of jacobian determinant
                As.col(idx) += ke.sign * nrm * w * sfp * dPhi.dot(dPhi) * dDetJ[idx]; // 1/S * (psi + grad_psi) * grad_phi^2 * ddetJ
            }
        }
    }
    return std::make_tuple(Av, As);
}
